// This file can be used to numerically test the dihedral-angle force code.
// To enable numerical testing: copy this file to the lammps "src" directory,
// uncomment the lines from "dihedral_table.cpp" which contain: 
// "#define  DIH_DEBUG_NUM"  and also to  "#undef NDEBUG,"
// and recompile.


#ifndef DIHEDRAL_TABLE_DEBUG_H
#define DIHEDRAL_TABLE_DEBUG_H
#include "domain.h"
#include "dihedral_table.h"
#include "dihedral_table_nd_mod.h"


namespace LAMMPS_NS {
namespace DIHEDRAL_TABLE_NS {

// This function compares the gradients computed in two different ways,
// and prints both results to the standard out.

inline void
PrintGradientComparison(DihedralTable &dih_tab,
                        double const *dphi_dx1, 
                        double const *dphi_dx2, 
                        double const *dphi_dx3, 
                        double const *dphi_dx4,
                        // We need to access dih_tab.domain, but domain is not a
                        // publically accessible member,so we pass it separately
                        Domain *domain, 
                        double *x1,
                        double *x2,
                        double *x3,
                        double *x4)
{
  double v_x[4][g_dim];
  double v_dphi_dx[4][g_dim];

  double vb12[g_dim];
  double vb23[g_dim];
  double vb34[g_dim];
  double n123[g_dim];
  double n234[g_dim];

  for (int d=0; d<g_dim; d++) {
    vb23[d]   = x3[d] - x2[d];
    v_x[0][d] = x1[d];
    v_x[1][d] = x2[d];
    v_x[2][d] = x3[d];
    v_x[3][d] = x4[d];
    v_dphi_dx[0][d] = dphi_dx1[d];
    v_dphi_dx[1][d] = dphi_dx2[d];
    v_dphi_dx[2][d] = dphi_dx3[d];
    v_dphi_dx[3][d] = dphi_dx4[d];
  }
  double L23 = sqrt(DotProduct(vb23, vb23));
  double deltax_d = 0.0002*L23; // (any reasonably small number will do)
  for (int i=0; i < 4; i++) {
    for (int d=0; d<g_dim; d++) {
      v_x[i][d] -= 0.5 * deltax_d;
      double phi1 = Phi(v_x[0], v_x[1], v_x[2], v_x[3], domain,
                        vb12, vb23, vb34, n123, n234);
      v_x[i][d] += deltax_d;
      double phi2 = Phi(v_x[0], v_x[1], v_x[2], v_x[3], domain,
                        vb12, vb23, vb34, n123, n234);
      double dphi = phi2 - phi1;
      // problem: If we are near the discontinuity in phi, then 
      //          phi1 and phi2 could differ by 2 pi.
      //          So we find the nearest image (due to periodicity)
      double dphi_round = TWOPI * ceil(dphi/TWOPI - 0.5);
      dphi = dphi - dphi_round; // This is difference between angle phi2 
                                // and the closest periodic image of phi1
      double dphi_dx_d = dphi / deltax_d;
      cerr << "[d_phi_dx" << i+1 << "[" << d << "]_theory - " 
           <<  "d_phi_dx" << i+1 << "[" << d << "]_numerical]  =  [("
           << v_dphi_dx[i][d] <<") - ("<< dphi_dx_d << ")]  =  "
           << v_dphi_dx[i][d] - dphi_dx_d;
      cerr << endl;
      // Assuming the bond-angles are reasonable, then the two gradients
      // should have a magnitude on the order of 1.0/L, (where L is a bond
      // length scale, approximately ~= L23).  However the -difference- between 
      // these two estimates should be less than 0.001/L (hopefully less).
      // Check that this is so.
    }
  }
} // PrintGradientComparison()

} //namespace DIHEDRAL_TABLE_NS
} //namespace LAMMPS_NS

#endif //#ifndef DIHEDRAL_TABLE_DEBUG_H

