/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   http://lammps.sandia.gov, Sandia National Laboratories
   Steve Plimpton, sjplimp@sandia.gov

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

/* ----------------------------------------------------------------------
   Contributing author: Andrew Jewett (jewett.aij at g mail)
------------------------------------------------------------------------- */
//    Due to popular demand, this file contains the extra code necessary 
// for calculating dihedral angles, and dihedral angle forces in 
// arbitrary dimensions higher than 3.
//    To test this, you can copy this file to the LAMMPS "src" directory.
// Then uncomment the line:  #include "dihedral_table_nd_mod.h"
// at the beginning of "dihedral_table.cpp", replace "Phi()" with "Phi_nd()",
// and recompile LAMMPS.  
//    The "Phi_nd()" defined in this file is a substitute for "Phi()" defined
// in "dihedral_table.h".  As of 2011/8/01, you can replace every instance 
// of "Phi()" in dihedral_table.cpp with "Phi_nd()", and the code should 
// still work (although this version might be slower).
//    Unfortunately, including this header file will not 
// make the rest of LAMMPS support higher dimensions.
//
// NOTE: WHEN g_dim>3, THIS CODE WILL APPEAR TO FAIL THE NUMERICAL TESTS IN
//       "dihedral_table_dbg.h"(that is, after changing Phi() to Phi_nd() there)
//        Attempting to change g_dim without rewriting the rest of LAMMPS 
//        will cause pointer array-bounds write errors, scrambled memory,
//        failed numerical tests, and crazy behavior.  Otherwise, this works.)


#ifndef LMP_DIHEDRAL_TABLE_ND_MOD_H
#define LMP_DIHEDRAL_TABLE_ND_MOD_H
#include "dihedral_table.h"


namespace LAMMPS_NS {
namespace DIHEDRAL_TABLE_NS {


// ---- ProjectOnto2dSubspace() and ----
// ---- FastProjectOnto2dSubspace() ----
// The following two functions compute the projection
// of vector "V" onto basis vectors "E1" and "E2"
// (not necessarily orthogonal) and stores the result in "dest".
// The version below is fast because it is assumed that many 
// of the dot-products have been computed already.
template<class _Real>
inline void FastProjectOnto2dSubspace(_Real const *E1,
                                      _Real const *E2,
                                      _Real *dest,
                                      _Real V_dot_E1,
                                      _Real V_dot_E2,
                                      _Real E1_sqd,
                                      _Real E2_sqd,
                                      _Real E1_dot_E2)
{
  // Formula used:
  //
  //             1      //                     \     /                     \  \.
  // dest = ----------- ||(<v,e1>-<v,e2><e1,e2>|e1 + |(<v,e1>-<v,e2><e1,e2>|e2|
  //        1-<e1,e2>^2 \\                     /     \                     /  /
  //
  // where v = V/|V|, e1 = E1/|E1|, e2 = E2/|E2| are unit vectors
  // and <e1,e2> denotes the dot-product.  Slightly more confusing to
  // read, the following formula is less expensive to calculate;
  // it only involves the original vectors, V, E1, and E2 (no square-roots):
  //
  //         //                           \    /                           \  \.
  // dest =C ||(<V,E1>|E2|^2-<V,E2><E1,E2>|E1 +|(<V,E2>|E1|^2-<V,E1><E1,E2>|E2|
  //         \\                           /    \                           /  /
  //
  //                       1
  //   and C = -------------------------
  //           |E1|^2*|E2|^2 - <E1,E2>^2
  //
  _Real one_over_outer_coeff =  E1_sqd * E2_sqd - (E1_dot_E2*E1_dot_E2);
  if (one_over_outer_coeff == 0.0)
    return;
  _Real outer_coeff = 1.0 / one_over_outer_coeff;

  _Real E1_coeff = outer_coeff * (V_dot_E1 * E2_sqd  -  V_dot_E2 * E1_dot_E2);
  _Real E2_coeff = outer_coeff * (V_dot_E2 * E1_sqd  -  V_dot_E1 * E1_dot_E2);
  for (int d=0; d < g_dim; ++d)
    dest[d] = E1_coeff * E1[d]  +  E2_coeff * E2[d];
}


// ProjectOnto2dSubspace() is more general than FastProjectOnto2dSubspace().
// But it is slower.  Use this version if you haven't already computed all the
// lengths and the dot-products beforehand.

template<class _Real>
inline void ProjectOnto2dSubspace(_Real const *V,
                                  _Real const *E1,
                                  _Real const *E2,
                                  _Real *dest)
{
  _Real E1_sqd    = DotProduct(E1, E1);
  _Real E2_sqd    = DotProduct(E2, E2);
  _Real V_dot_E1  = DotProduct(V, E1);
  _Real V_dot_E2  = DotProduct(V, E2);
  _Real E1_dot_E2 = DotProduct(E1, E2);

  FastProjectOnto2dSubspace(E1,
                            E2,
                            dest,
                            V_dot_E1,
                            V_dot_E2,
                            E1_sqd,
                            E2_sqd,
                            E1_dot_E2);
} //inline void ProjectOnto2dSubspace()



template<class _Real>
inline void SubtractFrom2dSubspace(_Real const *V,
                                   _Real const *E1,
                                   _Real const *E2,
                                   _Real *dest)
{
  ProjectOnto2dSubspace(V, E1, E2, dest);
  for (int d=0; d<g_dim; ++d)
    dest[d] = V[d] - dest[d];
} //inline void SubtractFrom2dSubspace()



// --------------------------------------------
// ------- Calculate the dihedral angle -------
// --------------------------------------------

inline 
double Phi_nd(double const *x1, //array holding x,y,z coords atom 1
	      double const *x2, // :       :      :      :        2
	      double const *x3, // :       :      :      :        3
	      double const *x4, // :       :      :      :        4
	      Domain *domain, //<-periodic boundary information
	      // The following arrays are of doubles with g_dim elements.
	      // (g_dim is a constant known at compile time, usually 3).
	      // Their contents is calculated by this function.
	      // Space for these vectors must be allocated in advance.
	      // (This is not hidden internally because these vectors
	      //  may be needed outside the function, later on.)
	      double *vb12, // will store x2-x1
	      double *vb23, // will store x3-x2
	      double *vb34, // will store x4-x3
	      double *n123, // will store normal to plane x1,x2,x3
	      double *n234) // will store normal to plane x2,x3,x4
{
  double phi;

  for (int d=0; d < g_dim; ++d) {
    vb12[d] = x2[d] - x1[d]; // 1st bond
    vb23[d] = x3[d] - x2[d]; // 2nd bond
    vb34[d] = x4[d] - x3[d]; // 3rd bond
  }

  //Consider periodic boundary conditions:
  //(Alas, the next 3 lines only make sense in 3-dimensions.
  // If this is a problem, just comment them out.)
  domain->minimum_image(vb12[0],vb12[1],vb12[2]);
  domain->minimum_image(vb23[0],vb23[1],vb23[2]);
  domain->minimum_image(vb34[0],vb34[1],vb34[2]);

  //--- Compute the normal to the planes formed by atoms 1,2,3 and 2,3,4 ---

  // (Note: I included the more general dihedral code that works in dimensions
  //        higher than 3.  The constant "g_dim" must be known at compile-time.

  if (g_dim == 3) 
  {                 
    // In 3-D, we can use the cross product to calculate the plane normals
    CrossProduct(vb12, vb23, n123);        // <- n123=vb12 x vb23
    CrossProduct(vb34, vb23, n234);        // <- n234=vb34 x vb23
    Normalize(n123);
    Normalize(n234);

    double cos_phi = -DotProduct(n123, n234);

    if (cos_phi > 1.0)
      cos_phi = 1.0;
    else if (cos_phi < -1.0)
      cos_phi = -1.0;

    phi = acos(cos_phi);

    if (DotProduct(n123, vb34) > 0.0) {
      phi = -phi;   //(Note: Negative dihedral angles are possible only in 3-D.)
      phi += TWOPI; //<- This insure phi is always in the range 0 to 2*PI
    }
  }
  else
  {
    // In higher dimensions, we calculate "phi" differently because the
    // plane "normals" n123 and n234 are difficult to define.   (Especially
    // true if all 4 points are coplanar, ie. when phi = 0 o 180 degrees).
    // Instead, it is more robust to calculate the angle between
    // "perp12on23" and "perp34on23".  Like vb12 and vb34, these two vectors
    // connect the first and last atoms  (atoms 1 and 4) onto the 
    // central-bond-axis (vb23).  But unlike vb12 and vb23, they are 
    // perpendicular this this axis.
    //    However we still need to compute n123 and n234 because the caller 
    // expects it, even though it's not needed for calculating phi in this case.
    // We do this below:
    SubtractFrom2dSubspace(vb34,  //Project vector connecting atoms 3 to 4
                           vb23,  //(vb34) onto the plane spanned by atoms
                           vb12,  //1,2,3 and subtract this from vb34,
                           n123); //<--and store the vector difference here
    SubtractFrom2dSubspace(vb12,  //Project vector connecting atoms 1 to 2
                           vb23,  //(vb12) onto the plane spanned by atoms
                           vb34,  //2,3,4 and subtract this from vb12,
                           n234); //<--and store the vector difference here
    for(int d=0; d<g_dim; d++) {
      //n123 and n234 currently point inward, towards eachother.  However,
      //as in 3-D, we want n123 and n234 to point in the direction of 
      //increasing phi. The sign of n123 is chosen so that movement of atom1
      //        in the direction of n123 increases phi ...
      //        ... AND the sign of n234 is chosen so that movement of atom4
      //        in the direction of n234 also increases phi
      n123[d] = -n123[d]; // ...SO, we must flip n123 and n234
      n234[d] = -n234[d]; //    to point outward, not inward
    }
    Normalize(n123);
    Normalize(n234);

    // Unfortunately, the code located in this section duplicates some of the 
    // code in DihedralTable::compute().  This is easy to fix but I'm lazy.

    //See figure in dihedral_table.cpp for a visual explanation of these vectors
    double proj12on23[g_dim];
    //    proj12on23[d] = (vb23[d]/|vb23|) * DotProduct(vb12,vb23)/|vb12|*|vb23|
    double proj34on23[g_dim];
    //    proj34on23[d] = (vb34[d]/|vb23|) * DotProduct(vb34,vb23)/|vb34|*|vb23|
    double perp12on23[g_dim];
    //    perp12on23[d] = v12[d] - proj12on23[d]
    double perp34on23[g_dim];
    //    perp34on23[d] = v34[d] - proj34on23[d]

    double dot123             = DotProduct(vb12, vb23);
    double dot234             = DotProduct(vb23, vb34);
    double L23sqr             = DotProduct(vb23, vb23);
    double inv_L23sqr         = 0.0;
    if (L23sqr != 0.0) 
      inv_L23sqr              = 1.0 / L23sqr;
    double dot123_over_L23sqr = dot123 * inv_L23sqr;
    double dot234_over_L23sqr = dot234 * inv_L23sqr;

    for (int d=0; d < g_dim; ++d) {
      proj12on23[d] = vb23[d] * dot123_over_L23sqr;
      proj34on23[d] = vb23[d] * dot234_over_L23sqr;
      perp12on23[d] = vb12[d] - proj12on23[d];
      perp34on23[d] = vb34[d] - proj34on23[d];
    }
    // Distance of atom 1 to the central axis:
    double perp12on23_len = sqrt(DotProduct(perp12on23, perp12on23));
    // Distance of atom 4 to the central axis:
    double perp34on23_len = sqrt(DotProduct(perp34on23, perp34on23));
    double numerator   = DotProduct(perp12on23, perp34on23);
    double denominator = perp12on23_len * perp34on23_len;
    double cos_phi;
    if (denominator != 0.0) 
      cos_phi = numerator / denominator;
    else
      cos_phi = 1.0;

    if (cos_phi > 1.0)
      cos_phi = 1.0;
    if (cos_phi < -1.0)
      cos_phi = -1.0;

    phi = acos(cos_phi);
  } // if (g_dim == 3) ... else

  return phi;
} // DihedralTable::Phi_nd()


} // namespace DIHEDRAL_TABLE_NS

} // namespace LAMMPS_NS


#endif //#ifndef LMP_DIHEDRAL_TABLE_ND_MOD_H
